/*\
 * MICE Column List View
 * Version 1.02
 * Copyright (C) 2000-2005 Colin Munro
 *
 * ListViewer.h - definitions of main objects for the CColumnListView object.
 *
 * Please see license.txt and readme.txt for more information on this product. Please
 * see readme.txt for more information if you wish to use this software in your own
 * product.
\*/

#ifndef __LISTVIEWER_H__
#define __LISTVIEWER_H__

#include <Invoker.h>
#include <View.h>
class CRow;
class CColumnListView;
class CListViewSorter;
class CColumnHeaderView;
class CColumnItemView;
#ifdef _BUILDINGLIB_
#include "ListViewerPrivate.h"
#endif // _BUILDINGLIB_

// Class to contain information on a single part of the list
class CField
{
public:
	CField();
	virtual ~CField();
private:
	friend CRow;
	friend CColumnItemView;
	
	void Attached(CField **parent);
	
	CField **parent;
	CField *last,*next;
	int iFieldIdentifier;
};

// Class to contain information about each row
class CRow
{
public:
	CRow(float height=16.0);
	virtual ~CRow();
	
	float Height(void) const;			// Height of this row
	float TotalHeight(void) const;		// Height of this rows children
	
	// Overrides for altering drawable stuff
	virtual void DrawExpander(BView *targetView,BRect rect);
	virtual bool DrawIcon(BView *targetView,BRect rect,float *width);	// return true if drawn, and the width of the icon
		// Note : DrawIcon() can use DrawBitmapAsync() on the targetView parameter, as the list view will call Sync() at the end
	virtual bool IsPointWithinExpander(BRect bigrect,BPoint point);
	virtual void ExpandOperation(bool bExpanding);			// Called before operation is done
	virtual void ExpandOperationComplete(bool bExpanding);	// Called after operation is done
	virtual bool IsExpandable(void);
	virtual void Update(void);	// Hook to update fields before a draw occurs
	
	// Functions for adding fields to the class
	void AddField(CField*,int iFieldIdentifier);
	
	bool IsExpanded(void) const;
	
	CRow* SuperItem(void);
	void SetExpanded(bool bExpanded);
	
	static float FindFontHeight(BFont *font);
	
protected:
	CRow* EachItemUnder(bool oneLevel,CRow* (*func)(CRow*,void*),void *data);

	CRow *last,*next;	// Self-maintaining linked list
private:
	friend CColumnListView;
	friend CColumnItemView;
	friend CListViewSorter;
	
	void Attached(CRow*,CRow**);
	void Unlink(void);
	void AddChild(CRow*);
	CField* FindField(int iFieldIdentifier);

	float fHeight;
	bool bExpanded;

	static CRow* EachItemUnder(CRow *rootrow,bool oneLevel,CRow* (*func)(CRow*,void*),void *data);
	
	CField *fields;		// Root of linked list of fields
	CRow **parent;		// Pointer to parent root
	CRow *owner;		// The actual parent row
	CRow *children;		// Root of linked list of child rows
	
	CRow *selection_last,*selection_next;
	bool IsSelected(void);
	void Deselect(void);
};

// Class to manage a single column in the list view
class CColumn
{
public:
	CColumn(float width,float minWidth,float maxWidth);
	virtual ~CColumn();

	// Width settings
	float Width(void) const;
	void SetWidth(float width);
	float MinWidth(void) const;
	float MaxWidth(void) const;
	
	typedef enum {slNone,slAscending,slDescending} seltype;
	
	// Overrides for altering drawable items
	virtual void DrawTitle(BView *targetView,BRect rect,seltype selected,bool bTransparent=false);
	virtual void DrawField(BView *targetView,BRect rect,CField *rowdata,bool bSelected,rgb_color highlight);
	virtual int CompareFields(CField *row1,CField *row2);
		// Note : DrawBitmapAsync() should be safe to use for the drawing functions
	
	// Show/hide column
	bool IsColumnVisible(void);
	
	// Position of column
	int Level(void) const;
private:
	friend CColumnListView;
	friend CColumnHeaderView;
	friend CColumnItemView;
	friend CListViewSorter;

	float fWidth,fMinWidth,fMaxWidth;
	int iFieldIdentifier,iLevel;
	bool bVisible;
	
	void Attached(CColumn**,int);
	void Unlink(void);
	CColumn *last,*next,**parent;
	
	static CColumn* FindColumn(CColumn*,int);
};

// Options for the CColumnListView constructor
#define LISTMODE_SINGLE		0		// Single selection
#define LISTMODE_MULTI		1		// Multiple selection

// Class that does most of the actual work drawing the list
class CColumnListView : public BView, public BInvoker
{
public:
	CColumnListView(BRect frame,char *name,uint32 resizingMode,uint32 flags,uint32 listMode);
	~CColumnListView();

	virtual void AllAttached(void);
	virtual void MessageReceived(BMessage *message);
	
	// Columns
	void AddColumn(CColumn *column,int iFieldIdentifier,char *name=NULL,int iLevel=0);
	void ShowColumn(CColumn *column,bool bVisible=true);
	void RemoveColumn(CColumn *column);	// column is NOT deleted. Deleting column will remove it from the view too, but not cause a redraw.
	virtual void Invalidate(CColumn *column);
	
	// Rows
	void AddRow(CRow *row,CRow *parent=NULL);
	void ExpandRow(CRow *row,bool bExpand=true);
	void RemoveRow(CRow *row);			// row is NOT deleted. Deleting row will remove it from the view too, but not cause a redraw.
	void MakeEmpty(void);
	virtual void Invalidate(CRow *row);
	virtual void Invalidate(CRow *row,int iFieldIdentifier);
	CRow* RowAt(int position,CRow *from=NULL,CRow *under=NULL);	// If "from" is NULL, the rows under "under" are used. Otherwise, it's done relative to "from" and "under" is ignored.
	bool IsSelected(CRow *row);

	void StartManyAdd(void);
	void EndManyAdd(void);
	
	// Sorting
	void DoSort(int iFieldIdentifier,bool bAscending=true);
	void DoSort(void) {DoSort(GetSortColumnSelection(),IsSortAscending());};
	bool IsSorting(void) const;
	void SetSorting(bool bSort);
	int GetSortColumnSelection(void) const;
	void SetSortColumnSelection(int iFieldIdentifier);
	bool IsSortAscending(void) const;
	void SetSortAscending(bool bAscending=true);

	// Searching
	CRow* SearchForNext(CRow *current,bool (*func)(CRow*,void*),void *data);
	CRow* FindNextItem(CRow *current,bool bIgnoreChildren=false);
	CRow* EachItemUnder(CRow *rootrow,bool oneLevel,CRow* (*func)(CRow*,void*),void *data);
	
	// Selection
	CRow* CurrentSelection(int index);
	void SetCurrentSelection(CRow *selection,bool bAdd);
	void ScrollTo(CRow *item);
	void ScrollToSelection(void);
	
	// Colours
	void SetHeaderBackground(rgb_color colour);
	rgb_color HeaderBackground(void) const;
	void SetListBackground(rgb_color colour);
	rgb_color ListBackground(void) const;
	void SetHighlight(rgb_color colour);
	rgb_color Highlight(void) const;

	// Faking keypresses
	void Key_Up(void);
	void Key_Down(void);
	void Key_Left(void);
	void Key_Right(void);
	
	// Other bits and bobs
	virtual void Invalidate(void);
	virtual void ContextMenuHook(BPoint point,uint32 buttons);	// Allows you to display a context menu for the item list, if desired
	virtual void MakeFocus(bool focused=true);
	virtual void SetFont(const BFont *font,uint32 properties=B_FONT_ALL);

	// Drag & Drop
	virtual BMessage* DragRequested(CRow *row);		// Return NULL if drags aren't allowed
	// Override MessageReceived() to process this

	// The following enumeration is a value passed along with the notification
	// message in the "nfaction" field as an int32. It indicates why the message
	// is occurring.
	typedef enum {nfSelection,nfSortOption,nfResize} NFACTION;
private:
	friend CColumnHeaderView;
	friend CColumnItemView;

	bool bSortAscending,bSortEnable;
	bool bAllowMultiple;

	CRow *rows;
	CColumn *columns;
	
	BScrollBar *verticalScroll;
	CColumnHeaderView *headers;
	CColumnItemView *items;
	
	CRow *selection;
	
	rgb_color highlight;
	
	void DoNotify(NFACTION);
	
	bool bFastAdd;
};

#endif // __LISTVIEWER_H__
